#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Converts a qli program generated by qli.py (a subset of Galil DMC) to an SVG file.

Copyriht Notice:
    This file is authored by Gianni Mariani <gianni@mariani.ws>. 27-May-2016.

    This file is part of qli_to_svg.

    qli_to_svg is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    qli_to_svg is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with qli_to_svg.  If not, see <http://www.gnu.org/licenses/>.
"""

import cmath
import math
import numpy
import svg.path as svg
from qli import qli_parser

import sys
from qli import value_type

from dataclasses import dataclass

# @dataclass
# class Hack:
#     path: str
# svg = Hack(svg_path)

def merge_dicts(*dict_args):
    """""Given any number of dicts, shallow copy and merge into a new dict,
    precedence goes to key value pairs in latter dicts.
    """
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result

EPLSILON = 1.e-6

def condition_floats(epsilon=EPLSILON, **kwds):
    """Returns a dict equivalent to the parameters where values less significant
    than epsilon are replaced with 0. This avoids cluttering the output file with 
    insignificant jibberish.
    """
    max_abs = 0
    for k, v in kwds.items():
        max_abs = max(max_abs, abs(v))
    
    result = {}
    for k, v in kwds.items():
        if abs(v) / epsilon < max_abs:
            result[k] = 0
        else:
            result[k] = v
    return result


SVG_HEADER = """
<svg width="{width:0.8g}" height="{height:0.8g}" xmlns="http://www.w3.org/2000/svg">

  <g transform="matrix({m00:0.8g},{m01:0.8g},{m10:0.8g},{m11:0.8g},{translatex:0.8g},{translatey:0.8g})">
"""

SVG_FOOTER = """
  </g>
</svg>
"""

def mat_rot(rotation):
    """Converts a rotation complex number (a complex number with unit size) to an
    equivalent 3x3 homogenous matrix."""
    return numpy.array([
                         [rotation.real, -rotation.imag, 0],
                         [rotation.imag,  rotation.real, 0],
                         [0, 0, 1]])

def mat_scale(scale):
    """Converts a scale complex number (real is x scale, imag is y scale) to an
    equivalent 3x3 homogenous matrix."""
    return numpy.array([
                         [scale.real, 0, 0],
                         [0, scale.imag, 0],
                         [0, 0, 1]])
    
def mat_trans(translation):
    """Converts a translation complex number (real is x translation, imag is y translation) 
    to an equivalent 3x3 homogenous matrix."""
    return numpy.array([
                         [1, 0, translation.real],
                         [0, 1, translation.imag],
                         [0, 0, 1]])

def svg_header(size, scale=1+1j, rotation=1+0j, extents=(0+0j,1+1j)):
    
    centre = extents[0] + (extents[1] - extents[0]) / 2
    
    # Operations - translate center to origin, scale, flip x axis, rotate, translate to new centre
    result_mat = (mat_trans(size / 2) @ mat_rot(rotation) @ mat_scale(-1+1j)
                  @ mat_scale(scale) @ mat_trans(-centre))
    
    return SVG_HEADER.format(
        **merge_dicts(
                {'width':size.real, 'height':size.imag},
                condition_floats(
                         m00=result_mat[0, 0], m10=result_mat[0, 1],
                         m01=result_mat[1, 0], m11=result_mat[1, 1]),
                condition_floats(
                        translatex=result_mat[0, 2], translatey=result_mat[1, 2])))
    
class BorderSpec(value_type.ValueSpec):
    VALUE_DELIMITER = ':'
    FIELDS = (value_type.ValueField('color', str, 'black', 'Color for objects'),
              value_type.ValueField('border_margin', float, 1., 'Border margin size factor'))
    
class BorderSpecList(value_type.ListSpec):
    LIST_SEPARATOR = ','
    LIST_TYPE = BorderSpec
    
class MarginSpec(value_type.ValueSpec):
    VALUE_DELIMITER = 'x'
    FIELDS = (value_type.ValueField('margin_width', float, 100., 'Margin width'),
              value_type.ValueField('margin_height', float, 100., 'Margin height'))
    
    def get(self):
        return complex(self.margin_width, self.margin_height)

class SvgOutputParams(value_type.ValueSpec):
    VALUE_DELIMITER = '|'
    FIELDS = (value_type.ValueField('oncolor', str, 'black', 'The color of the "on" path'),
              value_type.ValueField('offcolor', str, 'red', 'The color of the "off" path'),
              value_type.ValueField('line_width', float, 1., 'Bigger numbers means thicker lines'),
              value_type.ValueField('margin', MarginSpec, '50x50', 'Margin around image'),
              value_type.ValueField('width', float, 700.,
                                    'The image width (inside margin). Aspect ratio is maintained.'),
              value_type.ValueField('borders', BorderSpecList, 'blue:1',
                                    'Comma separated list of borders'))
        

SVG_PATH = """
    <path d="{path}" stroke="{color}" stroke-width="{stroke_width}" fill="none"/>
"""

def svg_path(path, color, stroke_width=5):
    return SVG_PATH.format(path=path, color=color, stroke_width=stroke_width)


class SvgOutContext(object):
    def __init__(self, pattern, params):
        self.pattern = pattern
        self.params = params
        
        extents = pattern.extents
        size = extents[1] - extents[0]
        self.bounding_box_delta = math.ceil((size.real + size.imag) * 0.02) * (1+1j)
        
        self.line_width = math.ceil((size.real + size.imag) * 0.001) * params.line_width

class SvgPath(object):
    def __init__(self, needle_state, path, alternate_color=None):
        self.needle_state = needle_state
        self.path = path
        self.alternate_color = alternate_color
        
        first = self.path[0]
        if not isinstance(first, svg.path.Move):
            self.path.insert(0, svg.path.Move(first.start))
        
        
    def svg(self, context):
        params = context.params
        color = params.oncolor if self.needle_state else params.offcolor
        if self.alternate_color:
            color = self.alternate_color
        return svg_path(self.path.d(), color, context.line_width)

        
class SvgPattern(object):
    def __init__(self, program, extents):
        self.program = program
        self.extents = extents
        self.svg_elements = []
        
    def add_bounding_box(self, color, offset=0j):
        box = svg.path.Path()
        extents = (self.extents[0] - offset, self.extents[1] + offset)
        box.append(svg.path.Line(extents[0], extents[0].real+1j*extents[1].imag))
        box.append(svg.path.Line(extents[0].real+1j*extents[1].imag, extents[1]))
        box.append(svg.path.Line(extents[1], extents[1].real+1j*extents[0].imag))
        box.append(svg.path.Line(extents[1].real+1j*extents[0].imag, extents[0]))
        self.append(SvgPath(False, box, color))
    
    def append(self, path):
        self.svg_elements.append(path)
        
    def write(self, f, context):
        pattern_size = self.extents[1] - self.extents[0]
        sizex = context.params.width
        scale = sizex / pattern_size.real
        sizey = scale * pattern_size.imag
        
        overall_size = complex(sizex, sizey) + 2 * context.params.margin.get()
        
        f.write(svg_header(overall_size,
                           complex(scale, scale),
                           rotation=cmath.rect(1, math.pi), extents=self.extents))
        
        for element in self.svg_elements:
            elem_str = element.svg(context)
            f.write(elem_str)
            
        f.write(SVG_FOOTER)
        

    def write_svg(self, f, params):
        pattern = self
        context = SvgOutContext(self, params)
        
        for border in params.borders:
            pattern.add_bounding_box(
                    border.color, context.bounding_box_delta * border.border_margin)
        
        pattern.write(f, context)        


class QliSvgExtentsExecutor(qli_parser.QliExecutor):
    """Computes the entents of a pattern.
    """
    def __init__(self, program):
        qli_parser.QliExecutor.__init__(self, program)
        self.axes = (0, 1)
        self.cur_loc = 0j
        self.min_extent = 0j
        self.max_extent = 0j
        self.relative = 0
        
    def coord(self, d1, d2):
        args = (d1, d2)
        return args[self.axes[0]] + 1j * args[self.axes[1]]
    
    def new_extent(self, loc):
        self.min_extent = complex(
                min(loc.real, self.min_extent.real), 
                min(loc.imag, self.min_extent.imag))
        self.max_extent = complex(
                max(loc.real, self.max_extent.real), 
                max(loc.imag, self.max_extent.imag))
    
    def doVectorMotion(self, index, command):
        self.axes = command.axes
            
    def doCircle(self, index, command):
        r = command.radius
        sa = command.start_angle * math.pi / 180.0
        ea = sa + command.angle_range * math.pi / 180.0
        sp = cmath.rect(r, sa)
        ep = cmath.rect(r, ea)
        delta = ep - sp
        next_loc = self.cur_loc + delta
        self.new_extent(next_loc)
        
        # Compute extents for rest of circle.
        saf = 2 * sa / math.pi
        eaf = 2 * ea / math.pi
        
        if eaf < saf:
            eaf, saf = saf, eaf
        
        saf = math.floor(saf)
        i = 1
        while i < 4 and saf + i < eaf:
            ex = cmath.rect(r, math.pi * 0.5 * (saf + i))
            self.new_extent(ex - sp + self.cur_loc)
            i += 1
            
        self.cur_loc = next_loc
    
    def doVectorPosition(self, index, command):
        next_loc = self.coord(command.d1, command.d2) + self.relative
        self.new_extent(next_loc)
        self.cur_loc = next_loc

    def doClearSequence(self, index, command):
        self.cur_loc = 0j

    def doVectorSequenceEnd(self, index, command):
        self.relative = self.cur_loc
        
    def get_extents(self):
        return (self.min_extent, self.max_extent)
    

class QliSvgExecutor(qli_parser.QliExecutor):
    
    def __init__(self, program):
        qli_parser.QliExecutor.__init__(self, program)
        self.axes = (0, 1)
        self.needle_on = True
        self.svg_current_path = svg.path.Path()
        self.program = program
        self.extents = QliSvgExtentsExecutor(program)
        program.execute(self.extents)
        
        self.start_loc = 0j
        self.cur_loc = self.start_loc
        self.pattern = SvgPattern(program, self.extents.get_extents())
        self.relative = 0
        
    def run(self):
        self.program.execute(self)
        self.end_path()
        
    def coord(self, d1, d2):
        args = (d1, d2)
        return args[self.axes[0]] + 1j * args[self.axes[1]]
        
    def doVectorMotion(self, index, command):
        self.axes = command.axes
        
        if False:
            sys.stderr.write("vector motion %s\n" % repr(command.groups))
            sys.stderr.write("axes = " + repr(command.axes) + "\n")
            
    def doCircle(self, index, command):
        r = command.radius
        sa = math.pi * (command.start_angle) / 180.0
        ea = sa + command.angle_range * math.pi / 180.0
        sp = cmath.rect(r, sa)
        ep = cmath.rect(r, ea)
        delta = ep - sp
        next_loc = self.cur_loc + delta
        mod_angle = (abs(command.angle_range) % 360)
        large_arc = mod_angle > 180
        
        sweep = command.angle_range > 0
        if True:
            arc = svg.path.Arc(self.cur_loc, complex(r, r), 0, large_arc, sweep, next_loc)
            self.svg_current_path.append(arc)
        else:
            self.add_debug_path(self.cur_loc, next_loc)
        self.cur_loc = next_loc
        
    def add_debug_path(self, cur_loc, next_loc):
        line = svg.path.Line(cur_loc, next_loc)
        self.svg_current_path.append(line)
    
    def doVectorPosition(self, index, command):
        next_loc = self.coord(command.d1, command.d2) + self.relative
        line = svg.path.Line(self.cur_loc, next_loc)
        self.svg_current_path.append(line)
        self.cur_loc = next_loc
    
    def doVectorSequenceEnd(self, index, command):
        self.end_path()
        self.relative = self.cur_loc
            
    def doClearSequence(self, index, command):
        self.cur_loc = self.start_loc
    
    def doNeedleOff(self, index, command):
        self.set_needle(False)
        
    def doNeedleOn(self, index, command):
        self.set_needle(True)
            
    def set_needle(self, state):
        # Previous motion should be generated.
        if self.needle_on != state:
            self.end_path()
        self.needle_on = state
        
    def end_path(self):
        if len(self.svg_current_path):
            self.pattern.append(SvgPath(self.needle_on, self.svg_current_path))
        self.svg_current_path = svg.path.Path()

    def doEnd(self, index, command):
        self.end_path()
        
    def get_svg_patttern(self):
        self.end_path()
        return self.pattern
    
        
        
        


"""
Converts from Statler Stitcher Gammill® QLI files to SVG format.

Gammill QLI files are a subset of the Galil DMC language. This script decodes
a small subset of the language and "runs" these programs to generate SVG files.

This script is capable of converting a single file or a whole directory of 
files.

Copyriht Notice:
    This file is authored by Gianni Mariani <gianni@mariani.ws>. 27-May-2016.

    This file is part of qli_to_svg.

    qli_to_svg is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    qli_to_svg is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with qli_to_svg.  If not, see <http://www.gnu.org/licenses/>.
"""

import argparse
import os
from qli import qli_parser
from qli import qli_svg
from qli import value_type
import re
import sys
import textwrap
import traceback
import subprocess

EXAMPLE_USAGE = """
To convert a single file from qli format to svg:
  {0} a-file.qli
  a-file.svg will be created.
  
To convert an entire directory hierarchy of qli files:
  {0} --recursive --continue_on_error a-directory
  All files named .qli will have a new file with the suffix replaced with .svg
  
To remove the border on the generated files:
  {0} --borders '' a-file.qli
  
To place the border on the shape extents:
  {0} --borders 'green:0' a-file.qli
  
To add two borders:
  {0} --borders 'blue:1,red:0' a-file.qli
  The second border will be drawn directly on the shape extents.
  
To additionally convert to to a jpeg image:
  {0} --raster_out png a-file.qli
  The raster out parameter is a comma separated list of a 4 tuple.
      type:<convert parameters>:<file format>:<out directory format>
  
Produce SVG with no margin or border and double wide lines:
  {0} --margin 0x0 --borders '' --line_width 2 a-file.qli
  The resulting file will have no border.
  
""".format(sys.argv[0])


DEFAULT_OUT_PATH_FORMAT='{path_name}'
DEFAULT_OUT_FORMAT='{base_name}.svg'
DEFAULT_FILE_EXTENSION='qli'

class RasterParams(value_type.ValueSpec):
    """Defines parameters for creating raster images using ImageMagick convert."""
    VALUE_DELIMITER = ':'
    FIELDS = (value_type.ValueField('raster_type', str, 'jpg', 'Raster output type'),
              value_type.ValueField('convert_args', str, '', 'Args to convert'),
              value_type.ValueField('file_name', str, '{base_name}.{raster_type}', 
                                    'Output file name pattern'),
              value_type.ValueField('out_dir', str, '{path_name}', 
                                    'The output directory for this type.'))

class RasterParamsList(value_type.ListSpec):
    LIST_SEPARATOR = ','
    LIST_TYPE = RasterParams

class ConvertParams(value_type.ValueSpec):
    """Conversion parameters class as a simple value type.
    """
    VALUE_DELIMITER = '|'
    FIELDS = (value_type.ValueField('recursive', bool, False, 
                                    'Perform recursive scan on directories.'),
              value_type.ValueField('continue_on_error', bool, False, 
                                    'Ignore parsing errors and omit the offending file.'),
              value_type.ValueField('out_dir', str, DEFAULT_OUT_PATH_FORMAT, 
                                    'Output directory format.'),
              value_type.ValueField('out_name', str, DEFAULT_OUT_FORMAT, 
                                    'Output file name format.'),
              value_type.ValueField('suffix', str, DEFAULT_FILE_EXTENSION, 
                                    'Suffix filter for recursive search.'),
              value_type.ValueField('print_progress', bool, False, 
                                    'Show progress of parsing.'),
              value_type.ValueField('raster_image', None, str, 
                                    'Add conversion to .'),
              value_type.ValueField('raster_out', RasterParamsList, '',
                                    'Comma separated list of raster types to generate')
              )

def readQliFile(filename):
    """Reads a QLI file and returns a tuple of (program, error).
    """
    f = None
    if filename == '-':
        f = sys.stdin
        sys.stderr.write("No file provided, reading from standard input:\n")
    else:
        f = open(filename, 'r')
        if not f:
            sys.stderr.write("failed to open %s" % filename)
            return None

    try:
        return qli_parser.QliProgram(filename, qli_parser.Qli(f)), None
    except qli_parser.QliSyntaxError as e:
        sys.stderr.write("File: %s %s" % (filename, e))
        return None, e

# group 1:path, 2:basename, 3:extension
FILE_NAME_RE = re.compile("(?:(.*[^/]|)/+)?([^/]*)(\\.[^\\./]*)")

def processQli(filename, prog, outname, svg_out_params=qli_svg.SvgOutputParams(), 
               convert_params=ConvertParams()):
    """Process a Qli program for generating SVG data.
    """
    try:
        converter = qli_svg.QliSvgExecutor(prog)
        if convert_params.print_progress:
            sys.stderr.write("%s: program extents %s\n" % (
                    filename, repr(converter.extents.get_extents())))
        converter.run()
        
        # Ensure the file is created.
        match = FILE_NAME_RE.match(outname)
        if not match:
            raise Exception("Unable to parse filename %s" % filename)
        
        dir, basename, extension = match.groups()
        if dir and not os.path.isdir(dir):
            os.makedirs(dir)
        if filename == "-":
            f = sys.stdout
        else:
            f = open(outname, 'w')
        converter.pattern.write_svg(f, svg_out_params)
        f.close()
        
        return True
    except:
        traceback.print_exception(*sys.exc_info())
        return False

def find_recursive(dir, suffix):
    """Returns a list of files matching the given filter."""
    suffix_matcher = re.compile(".*" + suffix + "$", re.IGNORECASE)
    result = []
    for (dirpath, dirnames, filenames) in os.walk(dir):
        for filename in filenames:
            if suffix_matcher.match(filename):
                result.append('/'.join((dirpath, filename)))
    return result

def error_exit(message):
    """Quick error exit."""
    sys.stderr.write(message)
    sys.stderr.write("\nEXIT\n")
    sys.exit(1)

def make_outfile_name(filename, dir_format, file_format, **kwds):
    """Creates an output filename from in input filename and format
    strings for directory and file and miscellaneous other parameters.
    """
    match = FILE_NAME_RE.match(filename)
    if not match:
        raise Exception("Unable to parse filename %s" % filename)
    
    local_params = {
        'path_name' : '.' if match.groups()[0] is None else match.groups()[0],
        'base_name' : match.groups()[1],
        'extension' : match.groups()[2]}
    
    params = qli_svg.merge_dicts(local_params, kwds)
    
    return '/'.join((dir_format.format(**params), file_format.format(**params)))


def generate_raster(outsvg_name, original_filename, convert_params):
    for raster_params in convert_params.raster_out:
        raster_name = make_outfile_name(original_filename, raster_params.out_dir,
                raster_params.file_name, raster_type=raster_params.raster_type)
        
        params = raster_params.convert_args.split(' ') if raster_params.convert_args else []
        
        command = [ "convert" ] + params + [ outsvg_name, raster_name]
        
        if subprocess.call(command):
            command_str = '"%s"' % '" "'.join(command)
            sys.stderr.write("Raster processing failed : %s\n" % command_str)

def main(argv):
    parser = argparse.ArgumentParser(
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description='Convert QLI pattern files to SVG.',
            epilog=textwrap.dedent(EXAMPLE_USAGE))
    
    svg_out_params = qli_svg.SvgOutputParams()
    svg_out_params.add_parser_args(parser)
    
    convert_params = ConvertParams()
    convert_params.add_parser_args(parser)

    parser.add_argument('inputs', metavar='<file>', type=str, nargs='*',
                   help='input files')
               
    args = parser.parse_args() 
    svg_out_params.set_parsed_args(args)
    convert_params.set_parsed_args(args)
    
    filenames = args.inputs if args.inputs else ['-']
    # Make a reversed copy, better for large lists.
    stack = filenames[::-1]
    
    try:
        filename = ""
        errors = []
        while stack:
            filename = stack.pop()
            if os.path.isdir(filename):
                if args.recursive:
                    stack = find_recursive(filename, args.suffix)[::-1] + stack
                    continue
                else:
                    return error_exit('cannot convert directory: use --recursive to walk directories')
            prog, e = readQliFile(filename)
            if e and args.continue_on_error:
                errors.append((filename, e))
            elif e:
                sys.stderr.write(str(e))
                return 1
            else:
                outname = outname=make_outfile_name(filename, args.out_dir, args.out_name)
                if not processQli(
                        filename=filename,
                        prog=prog,
                        outname=outname,
                        svg_out_params=svg_out_params,
                        convert_params=convert_params):
                    return 1
                generate_raster(outname, filename, convert_params)

    except KeyboardInterrupt as e:
        sys.stderr.write('\n\n')
        traceback.print_exc()
        # If the code gets stuck, it's good to know which file it was stuck on.
        sys.stderr.write('\nInterrupted at file: %s\n\n' % filename)
        return 1
    if errors:
        return error_exit('Some files failed conversion.')
    return 0


if __name__ == "__main__":
    exit(main(sys.argv))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Converts a .qli format file in a Qli program.

Copyriht Notice:
    This file is authored by Gianni Mariani <gianni@mariani.ws>. 27-May-2016.

    This file is part of qli_to_svg.

    qli_to_svg is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    qli_to_svg is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with qli_to_svg.  If not, see <http://www.gnu.org/licenses/>.
"""

import re
import sys
import traceback

def log(st, **kwds):
    sys.stderr.write(st + ": " + repr(kwds) + "\n")
    pass
#
# Error Exceptions

class InternalError(Exception):
    """An intenal parser consistency error was found.
    """
    pass

class QliSyntaxError(Exception):
    """A syntax error in parsed input was found.
    """
    def __init__(self, inputstr, pos, lineno=None, formaterr=None):
        self.inputstr = inputstr
        self.pos = pos
        self.lineno = lineno
        super(Exception, self).__init__(
            "Syntax error %s pos=%d line:%s%s" % (
                '' if not lineno else 'lineno=%s' % lineno,
                pos,
                inputstr,
                ' err=%s' % str(formaterr) if formaterr else ''))
        
class FormatError(Exception):
    pass

TERMINATOR = "(?:\\s*[;\n]|\r\n)"
NON_TERM_STR = "[^;\n]*"
NUMBER_RE = "\\s*([0-9.*/+\\-@\\[\\(\\]\\)A-Z]*)\\s*"
COMMA_NUMBER_RE = "," + NUMBER_RE
OPTIONAL_LT_NUMBER_RE = "(?:\\s*<" + NUMBER_RE + ")?"
OPTIONAL_GT_NUMBER_RE = "(?:\\s*>" + NUMBER_RE + ")?"
LABEL_RE = "\\s*([A-Za-z0-9]+)\\s*"

class Arg(object):
    """Defines components of a "Command".
    """
    def init(self, name, regex, group_count, required=True):
        self.name = name
        self.regex = regex
        self.group_count = group_count
        self.required = required
        
    def parse(self, groups):
        if len(groups) == 1:
            return groups[0]
        return ''.join(groups)
        
CRAXY1 = "(?:@RND\\[([+\\-]?[0-9]*\\.?[0-9]+)\\*S.\\])"
CRAXY2 = "(?:\\(([+\\-]?[0-9]*\\.?[0-9]+)\\*S.\\))"
class NumberArg(Arg):
    RE = re.compile('([+\\-]?[0-9]*\\.?[0-9]+)|%s' % '|'.join((CRAXY1, CRAXY2)))

    def __init__(self, name, required=True):
        self.init(name, NUMBER_RE, 1, required)
        
    def parse(self, groups):
        if len(groups) != 1:
            raise InternalError('Number parsing expecting 1 group for %s\n' % self.name)
        st = groups[0]
        if not st:
            if self.required:
                raise FormatError('Missing required arg %s\n' % self.name)
            return None
        match = self.__class__.RE.match(st)
        if not match:
            raise FormatError('Number format error for %s "%s"\n' % (self.name, st))
        for i in match.groups():
            if i:
                return float(i)
        raise FormatError('Number match failed for %s "%s"\n' % (self.name, st))

class CommaNumberArg(NumberArg):
    def __init__(self, name, required=True):
        self.init(name, COMMA_NUMBER_RE, 1, required)
        
class OptionalGtArg(NumberArg):
    def __init__(self, name, required=False):
        self.init(name, OPTIONAL_GT_NUMBER_RE, 1, required)
        
class OptionalLtArg(NumberArg):
    def __init__(self, name, required=False):
        self.init(name, OPTIONAL_LT_NUMBER_RE, 1, required)
        
class MatchAllArg(Arg):
    def __init__(self):
        self.init('text', NON_TERM_STR, 0, False)
        
class LabelArg(Arg):
    def __init__(self, name, required=True):
        self.init(name, LABEL_RE, 1, required)
        
class AxiiArg(Arg):
    def __init__(self, name, required=True):
        self.init(name, "([A-Za-z,]+)", 1, required)
 
    def parse(self, groups):
        if len(groups) != 1:
            raise InternalError('Axes parsing expecting 1 group for ' + self.name)
        st = groups[0]
        if not st:
            if self.required:
                raise FormatError('Missing required arg ' + self.name)
            return None
        if st == 'XY' or st == 'X,Y':
            return (0, 1)
        if st == 'YX' or st == 'Y,X':
            return (1, 0)
        if st == 'S':
            return (2,)
        
        raise FormatError('Unsupported axes defined ' + self.name)
        
class QueryArg(Arg):
    def __init__(self):
        self.init('query', "(?:\\s*(\\??))?\\s*", 1, False)
        
class BitMaskArg(Arg):
    def __init__(self, name):
        self.init(name, "\\s*([01]+)?", 1, True)

class Syntax(object):
    """ Defines a syntax. In the DMC language some constructs (REM) must
    occur at the beginning of the line.
    regex : The regular expression that matches the beginning of the construct.
    bol_only : indicates that this can only be matched a the beginning of a line.
    name : Optional name if the regex can't be used.
    terminator : What must be found to end the construct, usually a \n or ';'
    """
    def __init__(self, command, *args, **kwds):
        self.name = kwds.pop('name', command)
        terminator = kwds.pop('terminator', TERMINATOR)
        self.args = args
        regex = command
        for arg in args:
            regex = regex + arg.regex
        regex = regex + terminator
        self.re = regex
        self.bol_only = kwds.pop('bol_only', False)
        if kwds:
            raise InternalError("Extra arguments to Syntax constructor %r" % kwds)
        self.group_count = -1
        
    def parse_args(self, command):
        group_index = 0
        for arg in self.args:
            if arg.group_count:
                value = None
                try:
                    value = arg.parse(
                            command.groups[group_index + 1:(group_index + arg.group_count + 1)])
                except FormatError as e:
                    command.raiseFormatError(arg, e)
                setattr(command, arg.name, value)
                group_index += arg.group_count

class Command(object):
    """Command is an instance of a DMC "command". These specify how to parse
    the command from the input.
    """
    def __init__(self, groups, input_str, pos, lineno):
        self.groups = groups
        self.input_str = input_str
        self.pos = pos
        self.lineno = lineno
        self.__class__.SYNTAX.parse_args(self)
        self.commandName = "do" + self.__class__.__name__
        
    def raiseFormatError(self, arg, ex):
        raise QliSyntaxError(self.input_str, self.pos, lineno=self.lineno, formaterr=ex)
    
class RemComment(Command):
    SYNTAX = Syntax("REM", MatchAllArg(), bol_only=True)
        
class QuoteComment(Command):
    SYNTAX = Syntax("'|\"", MatchAllArg(), name='quote')  # A single ' or " to the end of line.
        
class NeedleOn(Command):
    SYNTAX = Syntax("NO NEEDLE ON", MatchAllArg(), name='needle_on')
    
class NeedleOff(Command):
    SYNTAX = Syntax("NO NEEDLE OFF", MatchAllArg(), name='needle_off')
    
class NoOperation(Command):
    SYNTAX = Syntax("NO", MatchAllArg())
        
class Label(Command):
    SYNTAX = Syntax('#', LabelArg('label'), name='label')

class ClearSequence(Command):
    SYNTAX = Syntax("CS")

class VectorMotion(Command):
    SYNTAX = Syntax("VM", AxiiArg('axes'))

class Circle(Command):
    SYNTAX = Syntax("CR", NumberArg('radius'), CommaNumberArg('start_angle'),
                    CommaNumberArg('angle_range'), OptionalLtArg('speed'),
                    OptionalGtArg('o'))

class VectorPosition(Command):
    SYNTAX = Syntax("VP", NumberArg('d1'), CommaNumberArg('d2'),
                    OptionalLtArg('speed'), OptionalGtArg('o'))
        
class WW(Command):
    SYNTAX = Syntax("WW", NumberArg('unk1'), CommaNumberArg('unk2'),
                    CommaNumberArg('unk3'))
        
class AN(Command):
    SYNTAX = Syntax("AN", NumberArg('unk1', required=False), CommaNumberArg('unk2', required=False),
                    CommaNumberArg('unk3', required=False))

class VectorSequenceEnd(Command):
    SYNTAX = Syntax("VE", QueryArg())

class Begin(Command):
    SYNTAX = Syntax("BG", AxiiArg('axes'))
        
class End(Command):
    SYNTAX = Syntax("EN", BitMaskArg('mask'))
        
class AfterMove(Command):
    SYNTAX = Syntax("AM", AxiiArg('axes'))
        
class AssignVariable(Command):
    SYNTAX = Syntax("([A-Za-z0-9]+)=", MatchAllArg(), name='assign')
        
class EmptyLine(Command):
    SYNTAX = Syntax("", name='empty')
        
class UnicodeBom(Command):
    # Some files contain a UTF8 BOM character which technically is illegal but some unicode
    # conversions incorrectly insert them. Let's eat them and ignore them.
    SYNTAX = Syntax("\xEF\xBB\xBF", name='unicode_bom', terminator='')
        
# A list of all regognized DMC commands. Add more as needed.
# Note that the QliRunner class must have a do<Class Name>() function for each element
# in this list.  When run, the functions are called.
# Because of the way we parse, we need the most commonly used commands at the front
# to make the parsing most efficient.
COMMANDS = [
    Circle,
    VectorPosition,
    RemComment,
    QuoteComment,
    NeedleOn,
    NeedleOff,
    NoOperation,
    Label,
    ClearSequence,
    VectorMotion,
    VectorSequenceEnd,
    Begin,
    AfterMove,
    End,
    WW,
    AN,
    AssignVariable,
    EmptyLine,
    UnicodeBom]


class ScannerMapper:
    """ Maps commands from regular expressions.
    """
    def __init__(self):
        self.regex = ""
        self.group_map = []
        self.compiled_re = None
        
    def add_term(self, command):
        next_regex = self.regex + ('|' if self.regex else '') + '(' + command.SYNTAX.re + ')'
        self.compiled_re = re.compile(next_regex)
        if not self.compiled_re or not self.compiled_re.groups > len(self.group_map):
            raise InternalError("Bad regex definition for %s" % command.__name__)
        # Extend the map to contain the new commands
        group_count = (self.compiled_re.groups - len(self.group_map))
        command.SYNTAX.group_count = group_count
        self.group_map.extend([command] * group_count)
        self.regex = next_regex
        
class QliScanner:
    """Scanner for a QLI string.
    """
    def __init__(self, commands):
        self.bol_mapper = ScannerMapper()
        self.mapper = ScannerMapper()
        self.next_group = 0
        for command in commands:
            self.bol_mapper.add_term(command)
            if not command.SYNTAX.bol_only:
                self.mapper.add_term(command)

    def scan(self, input_str, lineno=None):
        """Performs scanning of a QLI string. This may be a single line in a file
        or the entire contents of the file returning a list of commands.
        input_str: The string to parse.
        lineno: Optional line number to report for this string.
        """
        pos = 0
        bol = True
        result = []
        while pos < len(input_str):
            mapper = self.bol_mapper if bol else self.mapper
            match = mapper.compiled_re.match(input_str, pos)
            if not match:
                raise QliSyntaxError(input_str, pos, lineno=lineno)
            for index, group in enumerate(match.groups()):
                if group:
                    command = mapper.group_map[index]
                    group_count = command.SYNTAX.group_count
                    groups = match.groups()[index : index + group_count]
                    result.append(command(groups, input_str, pos, lineno))
                    bol = groups[0][-1] == '\n'
                    if bol and lineno != None:
                        lineno += 1
                    pos += len(groups[0])
                    break
        return result

SCANNER = QliScanner(COMMANDS)

class Qli:
    """Represents the parsed contents of a QLI file.
    """
    def __init__(self, inputvalue=None):
        self.commands = []
        if inputvalue:
            self.parse(inputvalue)
 
    def parse(self, inputvalue):
        """Parses a file like object."""
        if isinstance(inputvalue, str):
            self.commands.extend(SCANNER.scan(input))
        else:
            for lineno, line in enumerate(inputvalue):
                if line[-1] != '\n':
                    line += '\n'
                self.commands.extend(SCANNER.scan(line, lineno=lineno))


class QliRunner(object):
    """A QliProgram will use methods in this class to "execute" the program. Override
    these methods for application specific behaviour.
    """
    
    def __init__(self, program):
        self.program = program
        self.program_index = 0
        
    def run(self):
        self.program.execute(self)
        
    def doCircle(self, index, command):
        pass
    
    def doVectorPosition(self, index, command):
        pass
    
    def doRemComment(self, index, command):
        pass
    
    def doQuoteComment(self, index, command):
        pass
    
    def doNoOperation(self, index, command):
        pass
    
    def doNeedleOn(self, index, command):
        pass
    
    def doNeedleOff(self, index, command):
        pass
    
    def doLabel(self, index, command):
        pass
    
    def doClearSequence(self, index, command):
        pass
    
    def doVectorMotion(self, index, command):
        pass
    
    def doVectorSequenceEnd(self, index, command):
        pass
    
    def doBegin(self, index, command):
        pass
    
    def doAfterMove(self, index, command):
        pass
    
    def doEnd(self, index, command):
        pass
    
    def doWW(self, index, command):
        pass
    
    def doAN(self, index, command):
        pass
    
    def doEmptyLine(self, index, command):
        pass
    
    def doAssignVariable(self, index, command):
        pass
    
    def doUnicodeBom(self, index, command):
        pass

class ExecutionException(Exception):
    def __init__(self, message, e, tb):
        self.message = message
        self.e = e
        self.tb = tb
        
    def get_e(self):
        return self.e
    
    def get_tb(self):
        return self.tb
    
    def __str__(self, *args, **kwargs):
        return (self.message + '\n' 
                + '-' * 60 + '\n' 
                + str(self.e) 
                + '\n' + self.tb)
        

class QliProgram(QliRunner):
    """A 'compiled' version of the Qli.
    """
    def __init__(self, filename, qli):
        QliRunner.__init__(self, self)
        self.filename = filename
        self.qli = qli
        self.labels = {}
        # Collect labels by executing.
        self.execute(self)
        
    def execute(self, executor):
        # Execute this runner.
        try:
            executor.program_index = 0
            while executor.program_index < len(self.qli.commands):
                command = self.qli.commands[executor.program_index]
                index = executor.program_index
                executor.program_index += 1
                func = getattr(executor, command.commandName)
                func(index, command)
        except Exception as e:
            raise ExecutionException('file: "' + self.filename + '"', e, traceback.format_exc())
    
    def doLabel(self, index, command):
        self.labels[command.label] = index
    

class QliExecutor(QliRunner):
    
    def __init__(self, program):
        QliRunner.__init__(self, program)
        
        


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Allows for easily defining simple value type classes.

e.g.
import value_type
import sys
import textwrap
import argparse   

# Define a value type
class SomeValueType(value_type.ValueSpec):
    VALUE_DELIMITER = ':'
    FIELDS = (ValueField('color', str, 'black', 'Color for objects'),
              ValueField('border_margin', float, 1., 'Border margin size factor'))
  
# Define a list container that convers to and from a string.
class SomeValueTypeList(value_type.ListSpec):
    LIST_SEPARATOR = ','
    LIST_TYPE = SomeValueClass
    
# Defines a value type that is a composite of ValueSpec, ListSpec and other types
class Composite(value_type.ValueSpec):
    VALUE_DELIMITER = '|'
    FIELDS = (value_type.ValueField('some_value', SomeValueType, 'black:1', 'A SomeValueType value'),
              value_type.ValueField('some_list', SomeValueTypeList, 'black:1,red:1.5', 'A list of values'),
              value_type.ValueField('an_int', int, 11, 'An integer', 'd', value_type.no_operation),
              value_type.ValueField('a_bool', bool, False, 'A boolean', 'd', value_type.no_operation))


# Example main that 
def main(argv):
    parser = argparse.ArgumentParser(
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description='Example simple class type.')
    
    # Create a default value
    params = Composite(some_value=SomeValueType('green:33'))    
    # Prepare the ArgumentParser to parse for values in the params class.
    params.add_parser_args(parser)

    # Parse the args and put them in the params object
    args = parser.parse_args() 
    params.set_parsed_args(args)
    
    # Access the values.
    print(params.some_value)
    
    print(repr(params))


Copyriht Notice:
    This file is authored by Gianni Mariani <gianni@mariani.ws>. 27-May-2016.

    This file is part of qli_to_svg.

    qli_to_svg is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    qli_to_svg is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with qli_to_svg.  If not, see <http://www.gnu.org/licenses/>.
"""


import re

def no_operation(v):
    """Provided for types supported by a string format conversion."""
    return v

FALSE_REGEX = re.compile("^no|false|f|0+$", re.IGNORECASE)
def to_bool(v):
  return v.lower() in ("no", "false", "f", "0")

class ValueField(object):
    """Definition of a field.
    name: The field name. Must be a valid python identifier.
    type: The field type. Must accept a single string parameter for initialization.
    default_value: The default value for this field.
    description: Field description used for help in ArgumentParser.
    format_str: The format string to use for to string conversion.
    to_format: A value converter (or no_operation) for use with format_str.
    from_str_converter: If ommitted is the same as type, otherwise a convertsion function.
    """
    TYPE_CONV_MAP = {float: ('g', no_operation, float),
                     int: ('d', no_operation, int),
                     bool: ('d', no_operation, to_bool)}
    DEFAULT_CONVERION = ('s', str)
    def __init__(self, name, type, default_value, description, format_str=None, to_format=None,
                 from_str_converter=None):
        self.name = name
        self.type = type
        self.default_value = default_value
        self.description = description
        
        if format_str:
            self.format_str = format_str
        else:
            self.format_str = self.TYPE_CONV_MAP.get(type, self.DEFAULT_CONVERION)[0]
            
        if to_format:
            self.to_format = to_format
        else:
            self.to_format = self.TYPE_CONV_MAP.get(type, self.DEFAULT_CONVERION)[1]
            
        if from_str_converter:
            self.from_str_converter = from_str_converter
        else:
            self.from_str_converter = self.TYPE_CONV_MAP.get(type, (None, None, type))[2]


class ValueSpec(object):
    """A base class for value objects. The derived class muse define VALUE_DELIMITER 
    as the separator for field in a string and FIELDS containing an ordered list of 
    ValueField objects (order is important).
    """
    def __init__(self, value_str=None, **kwds):
        for field in self.__class__.FIELDS:
            setattr(self, field.name, kwds.get(field.name, field.default_value))
        
        self.set_as_str(value_str)
        
    def set_as_str(self, value_str):
        if not value_str:
            return
        for value, field in zip(
                value_str.split(self.__class__.VALUE_DELIMITER), self.__class__.FIELDS):
            setattr(self, field.name, field.from_str_converter(value))

    def __str__(self, *args, **kwargs):
        return self.__class__.VALUE_DELIMITER.join(
                    ('%' + field.format_str) % field.to_format(getattr(self, field.name))
                        for field in self.__class__.FIELDS)
        
    def __repr__(self, *args, **kwargs):
        return "%s(%r)" % (self.__class__.__name__, str(self))

        
    def add_parser_args(self, parser):
        for field in self.__class__.FIELDS:
            default_help = ''
            value = getattr(self, field.name)
            if value is not None:
                default_help = ' default: %s' % str(value)
            if field.type == bool:
                p = parser.add_mutually_exclusive_group(required=False)
                p.add_argument('--' + field.name, dest=field.name, action='store_true',
                               help=field.description + default_help)
                p.add_argument('--no-' + field.name, dest=field.name, action='store_false',
                               help='See --' + field.name)
                if not value is None:
                    p.set_defaults(**{field.name : value})
            else:
                parser.add_argument('--' + field.name, default=value, 
                        type=field.type, help=field.description + default_help)

    def set_parsed_args(self, args):
        for field in self.__class__.FIELDS:
            setattr(self, field.name, getattr(args, field.name))

class ListSpec(list):
    """Wrapper for a typed list that allows for conversion from a string to list
    with the delimiter LIST_SEPARATOR defined in derived classes.
    """
    def __init__(self, spec_str):
        if spec_str:
            self[:] = [self.__class__.LIST_TYPE(i)
                      for i in spec_str.split(self.__class__.LIST_SEPARATOR)]
        else:
            self[:] = []

    def __str__(self, *args, **kwargs):
        return self.__class__.LIST_SEPARATOR.join(str(i) for i in self)
        
    def __repr__(self, *args, **kwargs):
        return self.__class__.__name__ + '(%r)' % str(self)
    
